-- base methods for ant's enigma maps
-- (c) 2002-2003 Petr Machata/ant_39
-- Licensed under GPL v2.0 or above
--
-- 2003-01-07 -- special floor types and train support
-- 2003-01-11 -- multiples support (groups of doors, bolders etc.)
-- 2003-01-14 -- bugfixes and error reporting, some new map-creating funcs
-- 2003-02-09 -- 
--   many changes. Fixes of rubberband functions, fixes in multiples support, some interfaces changed
--   These changes require also fixes in several levels, but it's more logic this way. Besides this, it's
--   finally possible to call functions from init.lua or any functions declared like func(x, y, args),
--   for example oxyd(), fakeoxyd(), laser(), ...  and others.
--   While I was rewriting, I finally implemented multichar maps, which I planned since the beginning...
-- 2003-02-11 -- fixes in filling functions
-- 2003-02-12 -- cell() now returns last created object; functional-drawing functions accept table; and fixes

-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- MISCELANEOUS - -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- x,y should be numbers, although chars and strings are generally also the right choice, if you avoid ','
function getkey(x,y)
   return x..","..y
end

-- this couple of functions takes care of converting common tile coordinates to real coordinates of placed
-- actor. It depends on the given actor_mode, which is being declared like this:
-- cells["O"]=cell{parent=cells["+"], item={face="ac-blackball", attr={player=0}, actor=ACTOR_MODE}}
--
-- the meaning of actor mode is as follows:
-- value || a | x | y || meaning
-- ------++---+---+---++--------
--  0/nil|| 0 |  any  || this is not actor, but an item
--  1    || 1 | 0 | 0 || this is actor, centered in tile
--  3    || 1 | 1 | 0 || this is actor, centered vertically
--  5    || 1 | 0 | 1 || this is actor, centered horizontally
--  7    || 1 | 1 | 1 || this is actor, placed to left top corner of tile
--
function get_actor_x(x, actor_mode)
   if ((actor_mode == 3)or(actor_mode == 7)) then
      return x
   else
      return x + 0.5
   end
end

function get_actor_y(y, actor_mode)
   if ((actor_mode == 5)or(actor_mode == 7)) then
      return y
   else
      return y + 0.5
   end
end



-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- VISUAL MAP CREATION  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

TRANSPOSE_MAP = 1
NO_TRANSPOSE_MAP = not(TRANSPOSE_MAP)

-- width of key in map definition
-- CELL_KEY_WIDTH means how many characters occupy single map cell. In several maps I strongly
-- wanted this functions (cannonball for example), and after all here it is. Usable in maps with
-- wide variety of different surfaces.
CELL_KEY_WIDTH = 1
function cell_key_width(w)
   if ( (type(w) == "number") and (w>0) and (floor(w)==w) ) then
      CELL_KEY_WIDTH = w
      return w
   else
      print("warning: [ant.lua]: CELL_KEY_WIDTH has to be positive whole number")
      return -1
   end
end

-- this function tries to guess the width of key based on the most common width of keys in cell[key]
-- also there is possibility to get level as argument and try to divide its width by common level widths, 
-- such as 20, 39, ... this way it would be possible to get the value also... 
-- Anyway, I think first test is better.
function guess_cell_key_width(cellfuncs)
   local funcs = cellfuncs or cells or {} -- hope for a global cells
   local freq_table = {}

   for key,_ in funcs do
      local ktype = type(key)
      if (ktype == "string") then
	 local n = strlen(key)
	 freq_table[n] = (freq_table[n] or 0)+1
      end
   end

   local maxw,maxfreq = 0,0
   for w,freq in freq_table do
      if (freq>maxfreq) then
	 maxw = w
	 maxfreq = freq
      end
   end

   return cell_key_width(maxw)
end

-- each item:
--  + face - name of stone/floor/item object
--  + attr - set of attributes of given object
function cell_item(it)
   local n_it = {}
   n_it.face = (it.face or "")
   n_it.attr = (it.attr or {})
   return n_it
end

-- each cell
--  + stone - if present, the stone shall be added to defined position
--  + floor - the same, but for floor
--  + item  - the same, but for item
--   ++ actor - if item.actor~=0 is set, set_actor will be called instead of set_item.
function cell(structure)
   local cell0 = {}

   cell0.tag = (structure.tag or "?")
   cell0.floor = cell_item(structure.floor or {})
   cell0.stone = cell_item(structure.stone or {})
   cell0.item =  cell_item(structure.item or {})

   local parent =structure.parent
   local partype=type(parent)

   return function(x, y)
	     if (%partype=="nil") then
		-- no parent, do nothing!
	     elseif (%partype=="function") then
		-- parent is single function. let's call it!
		%structure.parent(x,y)
	     elseif (%partype=="table") then
		-- parent is table. look closer...
		for _,item in %parent do
		   if (type(item)=="function") then
		      -- item is function. call it!
		      item(x,y)
		   elseif (type(item)=="table") then
		      -- item is table. first item has to be of type "function"
		      -- the rest are the the function arguments
		      --
		      -- first construct global function __tmpfunc0, which calls defined
		      -- function from item[] and passes coordinates and arguments, like this:
		      --   item[1](x, y, item[2], item[3], ...)
		      -- the call it and finally erase it
		      -- well, it's dirty enough, but it works..
		      local counter = 0
		      local funccall = "__tmpfunc0 = function(uitem, ux, uy) "
		      --
		      for key,_ in item do
			 if (counter > 0) then funccall = funccall .. ", " end
			 funccall = funccall .. "uitem["..key.."]"
			 if (counter == 0) then funccall = funccall .. "(ux, uy" end
			 counter = counter + 1
		      end
		      funccall = funccall .. ") end"
		      --
		      dostring(funccall)
		      __tmpfunc0(item, x, y)
		      __tmpfunc0 = nil;
		   else
		      -- item is something else
		      print("warning: [ant.lua]: illegal parent item type:", type(item))
		   end
		end
	     else
		-- parent is something else
		print("warning: [ant.lua]: ilegal parent type:", %partype)
	     end

	     --map elements
	     local ret = nil

	     if (%structure.stone) then ret = set_stone(%cell0.stone.face, x, y, %cell0.stone.attr) end
	     if (%structure.floor) then ret = set_floor(%cell0.floor.face, x, y, %cell0.floor.attr) end
	     if (%structure.item) then
		if ((%structure.item.actor or 0) == 0) then
		   ret = set_item(%cell0.item.face, x, y, %cell0.item.attr)
		else
		   ret = set_actor(%cell0.item.face,
				   get_actor_x(x, %structure.item.actor),
				   get_actor_y(y, %structure.item.actor),
				   %cell0.item.attr)
		end
	     end

	     return ret
	  end
end

-- these return... voila... the sizes of the map
-- width - this is width of the widest row of map
function get_map_width(level)
   local levelw = 0

   for _,str in level do
      if strlen(str)>levelw then
	 levelw = strlen(str)
      end
   end

   return floor(levelw/CELL_KEY_WIDTH)
end

-- height is just number of lines
function get_map_height(level)
   return getn(level)
end

-- and size combined
function get_map_size(level)
   return get_map_width(level), get_map_height(level)
end

-- this draws the map to the position [x0,y0]
function draw_map(x0, y0, map, cellfuncs, swap_coords)
   local funcs = cellfuncs or cells or {}; --we hope that if nil is passed, global cells is defined.

   for y,str in map do
      local width = floor(strlen(str)/CELL_KEY_WIDTH);

      for x = 1,width do
	 local key = strsub(str, (x-1)*CELL_KEY_WIDTH+1, x*CELL_KEY_WIDTH)
	 local func = funcs[key]
	 local ftype = type(func)

	 if (ftype == "function") then
	    if (swap_coords) then
	       func(y+y0-1, x+x0-1)
	    else
	       func(x+x0-1, y+y0-1)
	    end
	 elseif (ftype == "nil") then
	    print("warning: [ant.lua]: function doesn't exist for map element '"..key.."'.")
	 else
	    print("warning: [ant.lua]: cell element '"..key.."' is not a function, it's "..ftype..".")
	 end
      end
   end
end

-- this prepares enigma world and draws given map
-- swap_coords is optional
function create_world_by_map(level, cellfuncs, swap_coords)
   local flavor = oxyd_default_flavor
   local levelw, levelh = get_map_size(level)

   if (swap_coords==TRANSPOSE_MAP) then
      temp, levelw, levelh = levelw, levelh, temp
   end

   create_world(levelw, levelh)
   oxyd_default_flavor = flavor or oxyd_default_flavor or "b"

   draw_map(0, 0, level, cellfuncs, swap_coords)
end

-- to fill each square of map with given function
function fill_world_func(fillfunc, x0, y0, w, h)
   local x0, y0 = (x0 or 0), (y0 or 0)
   local w, h = (w or 0), (h or 0)
   if w<=0 then w = level_width+w end
   if h<=0 then h = level_height+h end
   if (type(fillfunc)=="function") then fillfunc={fillfunc} end

   for _,func in fillfunc do
      for x=x0, x0+w-1 do
	 for y=y0, y0+h-1 do
	    func(x, y)
	 end
      end
   end
end

-- to draw border of map by given function
function draw_border_func(fillfunc, x0, y0, w, h)
   local x0, y0 = (x0 or 0), (y0 or 0)
   local w, h = (w or 0), (h or 0)
   if w<=0 then w = level_width+w end
   if h<=0 then h = level_height+h end
   if (type(fillfunc)=="function") then fillfunc={fillfunc} end

   for _,func in fillfunc do
      for x=x0,x0+w-1 do
	 func(x, y0)
	 func(x, y0+h-1)
      end

      for y=y0,y0+h-1 do
	 func(x0, y)
	 func(x0+w-1, y)
      end
   end
end

-- like set_stones, but calling a function
-- like: set_funcs(oxyd, {{1,2},{3,4},...})
function set_funcs(fillfunc, poslist)
   if (type(fillfunc)=="function") then fillfunc={fillfunc} end

   for _,func in fillfunc do
      for i,xy in poslist do
	 func(xy[1], xy[2])
      end
   end
end

-- like draw_stones, draw_floor and others, but calls a func
function draw_func(fillfunc, xy0, xystep, n)
   if (type(fillfunc)=="function") then fillfunc={fillfunc} end
   local x,y = xy0[1], xy0[2]

   for _,func in fillfunc do
      for i=1,n do
	 func(x, y)
	 x = x+xystep[1]
	 y = y+xystep[2]
      end
   end
end





-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- TRAINS/WORMS/WHATEVER FUNCTIONS  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

pathway_directions = {{ 1, 0}, { 0, 1}, {-1, 0}, { 0,-1}}
path = {}
path.constructor = {}

-- pathworks

-- tag shows whether there is a train or a rail - constructor sets it, destructor unsets it
-- context is ID of the track
function new_path_item(x, y, tag, context)
   local it = {}
   it.x, it.y = x, y
   it.tag = tag
   it.context = context

   return it
end

function add_path_item(x, y, tag, context)
   path[getkey(x,y)]=new_path_item(x,y,tag,context)
end

function add_path_train(x, y, context)
   add_path_item(x, y, nil, context)
end

function add_path_empty(x, y, context)
   add_path_item(x, y, not(nil), context)
end

-- constructors

function add_path_constructor(x, y, await, action, context)
   local cons = {}
   cons.x, cons.y = x, y
   cons.action    = action
   cons.await     = await
   cons.context   = context

   path.constructor[getn(path.constructor)+1] = cons
end

function move_constructor(x0, y0, awaittag, context)
   local found = nil

   for i,dir in pathway_directions do
      local x00, y00 = x0+dir[1], y0+dir[2]
      local key = getkey(x00, y00)

      if (path[key]) then
	 if ((path[key].context==context) and (path[key].tag==awaittag)) then
	    path[key].tag = not(path[key].tag)
	    found = i
	    break
	 end
      end
   end

   if (found) then
      local x1 = x0 + pathway_directions[found][1]
      local y1 = y0 + pathway_directions[found][2]
      return x1, y1
   else
      return x0, y0
   end
end

function move_constructors()
   local key, cons
   for key,cons in path.constructor do
      cons.x, cons.y = move_constructor(cons.x, cons.y, cons.await, cons.context)
      cons.action(cons.x, cons.y)
   end
end





-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- SPECIAL FLOOR TYPES  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

-- draws checker floor
-- sidex: horizontal size of checker square
-- sidey: vertical size of checker square
-- count: number of checker textures
-- items: table of checker textures [1,2,3,...]
-- x,y:   coordinates of checker field
function mcheckerfloor(x, y, sidex, sidey, offsetx, offsety, count, items)
   local x0, y0 = floor((x-offsetx+sidex)/sidex), floor((y-offsety+sidey)/sidey)
   local remainder = mod(x0+y0, count)
   items[remainder+1](x,y)
end

-- the checkerfloor parent
-- items: checkerfloor textures. Functions generated by cell({structure})
-- eg:
--   cells["."]=cell{floor={face="fl-metal"}}
--   cells[","]=cell{floor={face="fl-normal"}}
--   cells[" "]=cell{parent={{checkerfloor,{cells[","], cells["."]}}}}
-- optionally, at the end of table there may be additional checkerboard generating variables:
--   side - set size of checkerboard square
--   sidex - only set horizontal size (defaults to 1)
--   sidey - only set vertical size (defaults to 1)
--     cells[" "]=cell{parent={{checkerfloor,{cells[","], cells["."]; side=2}}}}
function checkerfloor(x, y, items)
   local count = getn(items)
   local remainder = mod(x+y, count)
   local sidex, sidey = 1, 1
   local offsetx, offsety = 0, 0

   if (items.side~=nil) then
      sidex = items.side
      sidey = sidex
   end

   if (items.sidex~=nil) then
      sidex = items.sidex
   end

   if (items.sidey~=nil) then
      sidey = items.sidey
   end

   if (items.offset~=nil) then
      offsetx = items.offset
      offsety = offsetx
   end

   if (items.offsetx~=nil) then
      offsetx = items.offsetx
   end

   if (items.offsety~=nil) then
      offsety = items.offsety
   end

   mcheckerfloor(x, y, sidex, sidey, offsetx, offsety, getn(items), items)
end

-- the random floor parent
--  every item may be:
--   texture -- parent function, for example cell[] function
--   occurence -- occurence factor for previous texture. Defaults to 1.
function randomfloor(x, y, items)
   local count = 0
   local total = 0 --total occurences
   local itemlist = {}

   for _,item in items do
      local titem = type(item)
      if (titem == "function") then
	 count = count + 1
	 itemlist[count] = {}
	 itemlist[count].func = item
	 itemlist[count].occf = 1
	 total = total + 1
      elseif (titem == "number") then
	 itemlist[count].occf = item
	 total = total + item-1
      end
   end

   local rand = total*random()
   local ctr  = 0

   for i=1,count do
      local move = itemlist[i].occf
      if ((rand>=ctr) and (rand<ctr+move)) then
	 itemlist[i].func(x,y)
      end
      ctr = ctr + move
   end
end

-- trains and pathworks support floor parents

-- path_constructor - add a generator of solid floor. This is the head of train.
-- opts[1] is context. It is number or string. Train only rides on the
--     paths with same context, as it has
-- opts[2] is constructing function. Usually it draws floor or something. It can also be one of cells[" "] functions.
function path_constructor(x, y, opts)
   add_path_constructor(x, y, not(nil), opts[2], opts[1])
end

-- path_destructor - add a desintegrator of solid floor. This is train's last wagon.
-- opts like in path_constructor
function path_destructor(x, y, opts)
   add_path_constructor(x, y, nil, opts[2], opts[1])
end

-- path_train - there already is body of train on this square
-- opts[1] is context, see path_constructor
function path_train(x, y, opts)
   add_path_train(x, y, opts[1])
end

-- path_empty - there is just path. Train can ride here eventually.
-- opts[1] is context, see path_constructor
function path_empty(x, y, opts)
   add_path_empty(x, y, opts[1])
end





-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- BOLDER/DOOR/ETC GROUPS  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

-- these are common parent function, much like checkerfloor or randomfloor
-- usage is as follows in example:
-- blocks = {};
-- cells["W"] = cell{parent={cells[" "],{add_multistone,{"st-block", blocks}}}}
--
-- this cell has two parents:
--   cells[" "], which probably is some common floor
--   {add_multistone,{"st-block", blocks}}, which is group parent function
--    add_multistone : add a stone to map and also to a group
--    "st-block"     : name of stone object to add
--    blocks         : LUA table that represents object group
--
-- Constructed group can be used later in the functions that work with it, like
--  send_group_message, add_rubber_bands, ...

-- add a stone to a group
-- first item in table is stone face
-- second is group table
-- third is object attributes
function add_multistone(x, y, args)
   local face, group, opts = args[1], args[2], args[3]
   local bc = getn(group)+1
   local bn = face..bc

   opts = opts or {}
   opts["name"] = bn

   set_stone(face, x, y, opts)
   local obj = enigma.GetNamedObject(bn)
   group[bc] = obj
end

-- add an actor to a group
-- first item in table is actor kind
-- second is group table
-- third is object attributes
function add_multiactor(x, y, args)
   local face, group, opts, actor_mode = args[1], args[2], (args[3] or {}), (args[4] or 1)
   local bc = getn(group)+1
   local bn = face..bc

   opts = opts or {}
   opts["name"] = bn

   set_actor(face, get_actor_x(x, actor_mode or 1), get_actor_y(y, actor_mode), opts)
   local obj = enigma.GetNamedObject(bn)
   group[bc] = obj
end

-- add the rubber band between all objects in group 1 and all objects in group 2
-- that is, each object from gr1 will be connected with each object from gr2
function add_rubber_bands(gr1, gr2, strength, length)
   for _,obj1 in gr1 do
      for _,obj2 in gr2 do
	 AddRubberBand(obj1, obj2, strength, length)
      end
   end
end

-- add pairs: 1st object from gr1 will be connected with 1st from gr2 and so on
-- requires same sized groups, of course...
function add_rubber_band_pairs(gr1, gr2, strength, length)
   for key1,obj1 in gr1 do
      AddRubberBand(obj1, gr2[key1], strength, length)
   end
end

-- send message to each object in group
function send_group_message(group, message, third)
   for _,item in group do
      enigma.SendMessage(item, message, third)
   end
end





-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- PUZZLE GENERATORS -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

-- use this function right like other parent function, for example checkerfloor or randomfloor
-- in contrast to common parent functions, its attribute should be a single table variable which
-- holds list of already registered puzzle cells. Usage:
--   puzzles = {};
--   cells["*"]=cell{parent={{puzzlecell, puzzles}}}
function puzzlecell (x, y, tab)
   tab[getkey(x,y)] = {}
   tab[getkey(x,y)].x = x
   tab[getkey(x,y)].y = y
end

-- this functions is called at the end of level file. It accepts
--  tab, which is a list of registered puzzle cells
--  generatorfunc, which is a cell[] function to add the floor or items to each registered cell
-- the function goes through all registered cells, observing their neighbors 
-- and adding right puzzle stones to right places
function render_puzzles (tab, generatorfunc)
   for _,val in tab do
      local x,y = val.x, val.y

      local up   = (tab[getkey(x, y-1)] ~= nil) or 0;
      local down = (tab[getkey(x, y+1)] ~= nil) or 0;
      local left = (tab[getkey(x-1, y)] ~= nil) or 0;
      local right= (tab[getkey(x+1, y)] ~= nil) or 0;

      if (generatorfunc) then
	 generatorfunc(x,y);
      end

      dostring("puz = PUZ_"..up..right..down..left)
      puzzle(x, y, puz)
   end
end
